#!/usr/bin/python3
# clean, sort, merge raw data
#
# output one huge table with all necessary fields for further processing/vizualization
#
# files are 'skyglow.log','tags' and trackfiles (containing 'unit-x' in name, where x is a number)
#
############################
# TODO
############################
# > parse moonmap and skyglowdict from skyglow Project sources 
# -> maybe integrate skyglow project as submodule
# > integrate with ST
#		> make Skyglow optional -> TEST
#		> ST-Filenames should include 'unit-X'!
#
# > how to process skylog error in merge()
# > ECOTRACK: include temperature data. Also humidity??
#
# > Speed Up! (takes 2min per 1mio lines)
#	1. Use Threads for
#	https://realpython.com/python-concurrency/#multiprocessing-version
#		> reading files (easy)
#		> merging (intermediate)
#		* parse chunks of trackData. call sort at end
#		* calc indices of data dependent on numThreads
#	2. rewrite animalmerge for loops
#	3. Consider using Pandas, NumPy, Cython, Pypy, Koalas, inline C++/C/Bash, sql (SQLAlchemy)

import sys, getopt, os, re, datetime, string, time, platform
from pprint import pprint as pp
from datetime import datetime as dt
import dataFilter

# Configuration of Time and paths
#####################################
# ecolux -- Track-start: 7.7.20 // Light start: 20.7.20 // Dark: 10.7.20 - 20.7.20
blox=[
("ecolux","./block-I/01-raw","./block-I/02-inspect","2020-07-21 00:00:00","2020-08-18 23:59:00"), 		# Block1: 21.7 - 18.8
("ecolux","./block-II/01-raw","./block-II/02-inspect","2020-09-15 00:00:00","2020-10-13 23:59:00"),		# Block2: 15.9 - 13.10
("ecotrack","/opt/virtualBoxVMs/share/block-I/01-raw","/opt/virtualBoxVMs/share/block-I/02-inspect","2019-10-14 00:00:00","2019-10-16 23:59:00"),
("ecolux","./test","./test","2020-09-14 00:00:00","2020-09-16 23:59:00"),
## -- add here -- ##
]

TIME_FMT='%Y-%m-%d %H:%M:00'
noTime = dt.fromtimestamp(0)
TAG_LEN=len("04B94A7F7288588022")
RULER=f"{'#'*100}\n"
rootPath = os.path.abspath( os.path.dirname( __file__ ) ) #uses cwd(), which is set to scriptdir at start

def initFileStructure():

	# trackfile
	##########
	global T_DELIM, TIMESTAMP_LEN, T_COL_MS, T_COL_X, T_COL_Y, T_COL_TAG, T_MINLEN, T_MAXLEN, T_NCOLS
	T_DELIM=';'
	TIMESTAMP_LEN=10
	#t_col_timestamp=0
	T_COL_MS=1
	T_COL_X=2
	T_COL_Y=3
	T_COL_TAG=4
	#t_col_signalstrength=5
	if PROJ == "ecolux":
		T_MINLEN = 38 # without newline.
		T_MAXLEN = 40 # ms column varies 1-3. Will be padded later for uniform len
		T_NCOLS = 6
	elif PROJ == "ecotrack": #ecotron temperature project 2019
		T_MINLEN = 36
		T_MAXLEN = 38
		T_NCOLS = 5

	# lightfile
	##########
	if hasLight():
		global L_DELIM, L_COL_MOON_REAL, L_COL_MOON_DMX, L_COL_SKYGLOW, L_MINLEN, L_MAXLEN, L_NCOLS
		L_DELIM='\t'
		L_COL_MOON_REAL=1
		L_COL_MOON_DMX=2
		L_COL_SKYGLOW=3
		L_MINLEN = 61 #2020-07-21 15:15:00 	0.0	0	[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
		L_MAXLEN = 76 #2020-07-21 15:09:00 	0.0	0	[15, 54, 4, 19, 0, 1, 5, 0, 40, 12, 123, 123]
		L_NCOLS = 4
	
	# animalfile
	###########
	global A_DELIM, A_MINLEN, A_MAXLEN, A_NCOLS, A_COL_SPECIES, A_COL_SPECIES_IND, A_COL_WEIGHT_WO_TAG_MG, A_COL_TAG
	if PROJ == "ecolux":
		A_DELIM='\t'
		A_MINLEN = 25
		A_MAXLEN = 30
		A_NCOLS = 4
		A_COL_SPECIES=0
		A_COL_SPECIES_IND=1
		A_COL_WEIGHT_WO_TAG_MG=2
		A_COL_TAG=3
	elif PROJ == "ecotrack":
		A_DELIM=';'
		A_MINLEN = 39
		A_MAXLEN = 44
		A_NCOLS = 7 #/w comments
		A_COL_SPECIES=1
		A_COL_SPECIES_IND=0
		A_COL_WEIGHT_WO_TAG_MG=2
		A_COL_TAG=5


	# out file
	###########
	global OUT_FMT, DELIM, COL_UID, COL_TAG
	OUT_FMT='%Y%m%d'
	DELIM=','
	COL_UID=4
	COL_TAG=8


#mlx (idx is also dmx-byte value)
moonMap = [ 1.409,
		 8.369,
		12.333,
		16.210,
		20.563,
		24.716,
		28.907,
		33.363,
		36.843,
		41.137,
		46.527,
		50.880,
		54.727,
		59.617,
		63.967,
		67.817,
		73.173,
		77.213,
		82.827,
		87.037,
		91.033,
		95.320,
		99.470,
		104.366,
		108.833,
		113.667, #25
		123.733,
		128.567,
		133.2,
		138.3,
		143.367,
		148.233,
		152.233,
		156.933,
		162.5,
		167.5,
		172.1,
		176.767,
		181.567,
		187,
		192.067,
		196.433,
		201.467,
		206.5,
		211.733,
		216.4,
		221.033,
		226,
		231.1,
		235.733,
		240.4, #50
		245.6,
		250.433,
		254.567,
		259.5,
		263.933,
		268.3,
		272.6, #274 == new max
		277.267,
		281.433,
		287.467, #60
	]

#Mapping skyglow dmx+nled to lux
skyglowDict = { 
			 15: 0.08,
			 54: 0.3,
			  4: 1,
			 19: 0.1,
			  0: 0,
			  1: 0.01,
			  5: 0.03,
			  0: 0,
			 40: 10,
			 12: 3,
			123: 30,
			123: 30,
}

def hasLight():
	return PROJ == "ecolux"

def getUID(fileName):
	""" extract and return unit-id from fileName """
	f=fileName.split('-')
	idx=f.index("unit")+1
	uid=int(f[int(idx)])
	return 1 if uid == 13 else uid

class Data:

	def __init__(self,animalFile, skyglowFile, trackFiles):

		self.animalData = self.getAnimalData(animalFile)

		self.trackData = []
		for file in trackFiles:
			self.trackData.extend( self.getTrackData(file) )

		self.trackData.sort()

		self.startTrack = self.getTrackTime( self.trackData[0] )
		self.endTrack = self.getTrackTime( self.trackData[-1] )

		if hasLight():
			self.lightData = self.getLightData(skyglowFile)
			self.startLight = self.getLightTime( self.lightData[0] )
			self.endLight = self.getLightTime( self.lightData[-1] )

		global OUT_FILE
		OUT_FILE = f"{PROJ}-data-{self.startTrack.strftime(OUT_FMT)}-{self.endTrack.strftime(OUT_FMT)}"

		pp(self.header())

	def getTrackData(self, fileName):
		""" get lines from all tracklogs, clean, merge and add/rm columns"""

		pp(f"processing: {fileName}")
		with open(INPUTDIR+os.sep+fileName) as f:

			lines = f.readlines()
			lines = self.clean(lines, T_MINLEN, T_MAXLEN, T_NCOLS, T_DELIM, self.getTrackTime)

			if not lines:
				pp("-> empty // non-usable")
				return []

		uid=getUID(fileName)
		idx_after_ms = TIMESTAMP_LEN + len('999') + len(DELIM)
		for idx, line in enumerate(lines):

			if PROJ in ["ecolux","ecotrack"]:
				# pad ms column with zeros to fit three digits
				cols=line.split(DELIM)
				ms_digits=len(cols[T_COL_MS])
				if ms_digits < 3:
					line = line[:TIMESTAMP_LEN+1] + '0'*(3-ms_digits) + line[TIMESTAMP_LEN+1:]

			trackTime = self.getTrackTime(line)
			date = trackTime.strftime('%Y-%m-%d')
			time = trackTime.strftime('%H:%M')

			x = int(cols[T_COL_X])
			y = int(cols[T_COL_Y])

			if x in (1,2) and y in (1,2):
				habitat=1
			elif x in (5,6) and y in (1,2):
				habitat=2
			elif x in (1,2) and y in (5,6):
				habitat=3
			elif x in (5,6) and y in (5,6):
				habitat=4
			else:
				habitat=0

			# add columns (date,time,uid,habitat)
			line = line[:idx_after_ms] + DELIM + date + DELIM + time + DELIM + str(uid) + DELIM + str(habitat) + line[idx_after_ms:]

			if PROJ == "ecolux":
				line = line[:line.rindex(DELIM)] # remove signalstrength column

			lines[idx] = line

		assert (lines), "trackData: Not valid. Check times / formatting."
		return lines

	def getLightData(self,file):
		
		with open(file) as f:
			lines = f.readlines()
			lines = self.clean(lines, L_MINLEN, L_MAXLEN, L_NCOLS, L_DELIM, self.getLightTime)
			pattern = re.compile(r"\s\t") # skylog used to have an xtra space (changed 200915) TODO remove
			lines = [re.sub(pattern, "\t", x) for x in lines ]

		assert (lines), "skyglow: Not valid. Check times / formatting."
		return lines

	def getAnimalData(self,file):
		
		with open(file) as f:
			lines = f.readlines()

		lines = self.clean(lines, A_MINLEN, A_MAXLEN, A_NCOLS, A_DELIM)		
		#TODO: check for duplicate tags and print warning / ignore?
		
		assert (lines), "tags: Not valid. Check times / formatting."
		return lines

	def clean(self, lines, minLen, maxLen, nCols, sep, timeFunc=None):
		""" remove trailing newline, empty and comments and remove lines before startTime or after endTime"""
		lines = [l.strip() for l in lines]
		lines = [l for l in lines if minLen <= len(l) <= maxLen] # ecolux: ms are variable in length (0-999), SchrebaTron: ms is padded with zeros
		lines = [l for l in lines if len(l.split(sep)) == nCols]
		lines = [l.replace(sep,DELIM) for l in lines]
		lines = list(filter(lambda q: q and q[0] != '#', lines))
		if timeFunc:
			lines = list(filter(lambda q: startTime <= timeFunc(q) < endTime, lines)) #only between start and endTime

		return lines

	def parseLightLine(self, line, uid):
		if not hasLight():
			return

		cols=line.split(DELIM)
		moon_real = float(cols[L_COL_MOON_REAL])*1000 #convert to mLux
		moon_dmx = int(cols[L_COL_MOON_DMX])
		moon_eco = moonMap[moon_dmx]

		#int list from string format like '[1, 2, 3, 4, 5]'
		skyglowList = cols[L_COL_SKYGLOW:]
		skyglowList[0] = skyglowList[0][1:] 	#remove '['
		skyglowList[-1] = skyglowList[-1][:-1]	#remove ']'
		skyglow = skyglowDict.get(int(skyglowList[uid-1]),'')

		return (moon_real, moon_eco, skyglow)

	def merge(self):
		
		self.merge_animals()
		if hasLight():
			self.merge_light()

	def merge_animals(self):
		""" merge AnimalData into TrackData on TagID """
		pp("merging animal data into track data. This might take a while. Go grab a Coffee or do some pushups!")

		#TODO SPEED UP!!! 

		# ttags_not_found=[]
		notFound=0
		for idx, tLine in enumerate(self.trackData):
			found=False
			ttag = tLine.split(DELIM)[COL_TAG]
			for aLine in self.animalData:
				aLine = aLine.split(DELIM)
				atag = aLine[A_COL_TAG]

				if ttag == atag:
					found=True
					aLine = DELIM + aLine[A_COL_SPECIES_IND] + DELIM + aLine[A_COL_SPECIES] + DELIM + aLine[A_COL_WEIGHT_WO_TAG_MG]
					self.trackData[idx] += aLine
					break;

			if not found:
				self.trackData[idx] += f'{DELIM}NA{DELIM}NA{DELIM}NA'
				notFound+=1
				# ttags_not_found.append(ttag)

		pp(f"Didn't find Tags for {notFound} tracking events!")

	def merge_light(self):
		"""merge LightData into TrackData on Time"""
		pp("merging light data into track data")

		track_idx=0
		light_idx=0
		lightTime = self.startLight
		trackTime = self.startTrack
		
		while( trackTime < lightTime ):
			self.trackData[track_idx] += f'{DELIM}NA{DELIM}NA{DELIM}NA'
			track_idx+=1
			if( track_idx >= len(self.trackData) ):
				pp("skyglow.log starts after track.log...")
				return
			trackTime = self.getTrackTime(self.trackData[track_idx])

		pp(f"start merge @ {trackTime} -- idx {track_idx}")

		for trackLine in self.trackData[track_idx:]:
			trackTime = self.getTrackTime(trackLine)
			
			uid = int(trackLine.split(DELIM)[COL_UID])

			while lightTime < trackTime and light_idx < len(self.lightData) - 1: # -> minute precision
				light_idx+=1
				lightTime = self.getLightTime(self.lightData[light_idx])

			#add light cols
			moon_real, moon_eco, skyglow = self.parseLightLine(self.lightData[light_idx],uid)
			trackLine += DELIM + "{:.3f}".format(moon_real) + DELIM + "{:.3f}".format(moon_eco) + DELIM + str(skyglow)

			self.trackData[track_idx] = trackLine
			track_idx+=1

	def getLightTime(self,line):
		if not hasLight():
			return

		cols = line.strip().split(DELIM)
		sTime = cols[0].strip()
		if "Error" in sTime:
			return noTime
		
		lightTime = dt.strptime(sTime,TIME_FMT)

		#convert to UTC from local BerlinTZ and respect DST sun 25 october 3AM -> -1 hour = 2AM again
		winterTime = dt.strptime("2020-10-25 03:00:00",TIME_FMT)
		hours= ( 2 if lightTime < winterTime else 1 )

		lightTime -= datetime.timedelta(hours)
		return lightTime


	def getTrackTime(self,line):
		""" convert timestamp string to datetime """
		try:
			timestamp=int(line[:TIMESTAMP_LEN])
		except ValueError as e:
			print("Caught:",e)
			print("Presumably binary data in txt file... or some other corruption")
			return noTime

		return dt.fromtimestamp(timestamp)

	def header(self):
		h= f"# {PROJ} data\n" \
		f"{RULER}" \
		f"#     len(Track): {len(self.trackData)}\n" \
		f"#     startTrack: {self.startTrack}\n" \
		f"#       endTrack: {self.endTrack}\n" \
		f"#    len(animal): {len(self.animalData)}\n"
		if hasLight():
			h += 	f"#     len(Light): {len(self.lightData)}\n" \
					f"#     startLight: {self.startLight}\n" \
					f"#       endLight: {self.endLight}\n"
		h += RULER
		h += f"#timestamp, ms, date, time, unit, habitat, x, y, tag, species, speciesnumber, weight_without_tag[mg]{', moon_real[mLux], moon_eco[mLux], skyglow[Lux]' if hasLight() else ''}\n"
		h += RULER
		return h

	def write(self):
		os.makedirs(OUTPUTDIR, exist_ok=True)
		os.chdir(OUTPUTDIR)
		pp("write to " + OUT_FILE)
		with open(OUT_FILE, "w") as f:
			f.writelines(self.header())
			self.trackData=map(lambda x:x+'\n', self.trackData)
			f.writelines(self.trackData)

		cmd_compress=f"tar -C {OUTPUTDIR} -zcvf {OUT_FILE}.tgz {OUT_FILE}" # ">/dev/null 2>/dev/null"
		os.system(cmd_compress)

def xtract():
	for root, dirs, files in os.walk(INPUTDIR): #walk recursively
		for file in files:
			if file.endswith(".tgz") or file.endswith(".tar.gz"):
				print(f"xtracting {file}\n")
				cmd_xtract=f"tar -zxvf {INPUTDIR}{os.sep}{file} -C {INPUTDIR}"
				os.system(cmd_xtract)
		return

def getFileList():

	animalFile, skyglowFile, trackFiles = '','',[]
	print( "Get files." )
	for root, dirs, files in os.walk(INPUTDIR): #walk recursively
		for file in files:
			if "unit" in file and not file.endswith(".tgz") and not file.endswith(".tar.gz"):
				trackFiles.append(file)
			elif file == "tags":
				animalFile = INPUTDIR+os.sep+file
				print(f"Found animalFile: {animalFile}")
			elif hasLight() and file == "skyglow.log":
				skyglowFile = INPUTDIR+os.sep+file
				print(f"Found logfile: {skyglowFile}")
			else: 
				pp(f"Ignoring ({file})")

	if not trackFiles:
		sys.exit("No tracking data found. (filename containing unit-X, where X is a number)")
	if hasLight() and not skyglowFile:
		sys.exit(f"Couldn't find 'skyglow.log'")
	if not animalFile:
		sys.exit(f"Couldn't find 'tags'")

	return animalFile, skyglowFile, trackFiles


def make_rel_abs_path(path):
	path = os.path.expanduser(path)
	if not os.path.isabs(path):
		path = os.path.abspath(rootPath+os.sep+path)
	return os.path.normpath(path)

def main():

	initFileStructure()
	xtract()
	
	animalFile, skyglowFile, trackFiles = getFileList()
	data = Data(animalFile,skyglowFile,trackFiles)
	data.merge()
	data.write()

	dataFilter.main(OUT_FILE)

	if( platform.system() == "Linux" ):
		os.system(rootPath+"/updateGitIgnore.sh")

if __name__ == "__main__" :
	known_projects=["ecotrack","ecolux","schrebatron"]
	print(f"	Process raw data for which block?\n {RULER}")
	for idx, b in enumerate(blox):
		print(f"    {idx+1}) Project {b[0]}")
		print(f"        from: {b[3]}")
		print(f"          to: {b[4]}")
		print(f"          in: {b[1]}")
		print(f"         out: {b[2]}\n")
	try:
		n=int(input())-1
		PROJ = blox[n][0].lower()
		INPUTDIR = make_rel_abs_path( blox[n][1] )
		OUTPUTDIR = make_rel_abs_path( blox[n][2] )
		startTime = dt.strptime(blox[n][3], TIME_FMT)
		endTime = dt.strptime(blox[n][4], TIME_FMT)

		if PROJ not in known_projects:
			raise Exception(f"Unknown Project '{PROJ}'")

	except Exception as e:
		print(f"Error: {e}\n")
		print("You can add a configuration by adding a line to the 'blox' structure. It has following form:\n")
		print("			(Project, inDir, outDir, start, end)\n" )
		print(f"> Project must be in {known_projects}" )
		print(f"> dirs can be relative. eg: ('./data','../data','~/data')")
		print(f"> times must have Format like '{dt.now().strftime(TIME_FMT)}'" )
		exit()

	start = time.time()
	main()
	end = time.time()
	print(f"Took {end - start}")
